package ch.fritteli.labyrinth.generator.renderer.pdffile;

import ch.fritteli.labyrinth.generator.model.Labyrinth;
import ch.fritteli.labyrinth.generator.renderer.Renderer;
import ch.fritteli.labyrinth.generator.renderer.pdf.PDFRenderer;
import io.vavr.control.Option;
import io.vavr.control.Try;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.NoSuchElementException;
import lombok.NonNull;
import lombok.extern.slf4j.Slf4j;

@Slf4j
public class PDFFileRenderer implements Renderer<Path> {

    @NonNull
    private static final PDFRenderer PDF_RENDERER = PDFRenderer.newInstance();
    @NonNull
    private Option<Path> targetFile;

    private PDFFileRenderer() {
        this.targetFile = Try
                .of(() -> Files.createTempFile("labyrinth_", ".pdf"))
                .onFailure(ex -> log.error("Unable to set default target file.", ex))
                .toOption();
    }

    @NonNull
    public static PDFFileRenderer newInstance() {
        return new PDFFileRenderer();
    }

    public boolean isTargetFileDefinedAndWritable() {
        return this.targetFile
                .map(Path::toFile)
                .exists(File::canWrite);
    }

    @NonNull
    public PDFFileRenderer setTargetFile(@NonNull final Path targetFile) {
        this.targetFile = Option.of(targetFile);
        return this;
    }

    @NonNull
    @Override
    public Path render(@NonNull final Labyrinth labyrinth) {
        if (!this.isTargetFileDefinedAndWritable()) {
            try {
                Files.createFile(this.targetFile.get());
            } catch (IOException | NoSuchElementException e) {
                throw new IllegalArgumentException("Cannot write to target file.", e);
            }
        }
        final ByteArrayOutputStream bytes = PDF_RENDERER.render(labyrinth);
        final Path outputFile = this.targetFile.get();
        try {
            bytes.writeTo(Files.newOutputStream(outputFile));
        } catch (IOException e) {
            log.error("Failed writing to file " + outputFile.normalize(), e);
        }
        return outputFile;
    }
}
